/* Author: Beno Gorancic
   Date:   04/11/2006

   Usage: ./grep n searchstring

   Where n is the number of threads and
   searchstring is the text to search for.
   n must be > than 0.  A maximum of 1 thread
   per file can be allocated.
   
   note: copy the file to the directory you want to search
*/

#include <pthread.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>

#define LINE_LENGTH 512

typedef struct
{
   char *filename;
   char *string_to_search;
} thread_data;

DIR *opendirectory(const char *dir);
struct dirent *readdirectory(DIR *dir, int num_of_threads, char *string_to_search);
int closedirectory(DIR *dir);
FILE *openFile(char *filename);
char searchstring(char *filename, char *string_to_search);
void *threadsearch(void *arg);

int main(int argc, char *argv[])
{
   DIR *dfd;
   struct dirent *dptr;

   if( (argc != 3) || (atoi(argv[1]) < 1) )
   {
      printf("Usage: %s n searchstring\nwhere n > 0\n", argv[0]);
      exit(0);
   }

   dfd = opendirectory(".");
   dptr = readdirectory(dfd, atoi(argv[1]), argv[2]);
   closedirectory(dfd);

   exit(0);
}

DIR *opendirectory(const char *dir)
{
   /* open a directory stream */

   DIR *dfd;

   if((dfd = opendir(dir)) == NULL)
   {
      perror("'open directory' error");
      exit(1);
   }

   return dfd;
}

struct dirent *readdirectory(DIR *dir, int num_of_threads, char *string_to_search)
{
   struct stat dir_stat;
   struct dirent *dptr;
      
   int x=1;
   int y, res;
   
   thread_data *my_data;
   pthread_t *thread;
   
   my_data = malloc(num_of_threads*sizeof(thread_data));
   thread = malloc(num_of_threads*sizeof(pthread_t));

   /* read in filenames */
   while((dptr = readdir(dir)) != NULL)
   {
      stat(dptr->d_name, &dir_stat);

      /* skip sub directory */
      if(S_ISDIR(dir_stat.st_mode))
         continue;

      /* skip '.', '..' & 'grep' within the directory */
      if(!((strcmp(dptr->d_name, ".") == 0) || (strcmp(dptr->d_name, "..") == 0) || (strcmp(dptr->d_name, "grep") == 0)))
      {
         if(x < num_of_threads)
         {
            my_data[x].filename = dptr->d_name;
            my_data[x].string_to_search = string_to_search;

            if((res = pthread_create(&thread[x], NULL, threadsearch, &my_data[x])) != 0 )
            {
               perror("'create thread' error");
               exit(1);
            }
            x++;
         }
         else
            searchstring(dptr->d_name, string_to_search);
      }
   }

   /* join threads before going any further */   
   for(y=1; y < x; y++)
      pthread_join(thread[y], NULL);

   return 0;
}                 

int closedirectory(DIR *dir)
{
   /* close the stream */
   
   if (closedir(dir) == -1)
   {
      perror("'close directory' error");
      exit(1);
   }

   return 0;
}

FILE *openFile(char *filename)
{
   /* open a FILE stream to filename with given mode */

   FILE *file;

   if((file = fopen(filename, "rb")) == NULL)
   {
      perror("'read file' error");
      exit(1);
   }                     

   return file;
}

char searchstring(char *filename, char *string_to_search)
{
   /* searches a file for text */

   FILE *file;
   int line=0;
   unsigned int x, y, u;
   char buffer[LINE_LENGTH];

   file = openFile(filename);

   while(1)
   {
      y = 0;
      u = 0;

      fgets(buffer, LINE_LENGTH, file);
      line++;

      if(feof(file))
         break;

      for(x = 0; x < strlen(buffer); x++)
      {
         if(buffer[y] == string_to_search[u])
         {
            y++;
            u++;

            if(u == strlen(string_to_search))
               printf("%s Line %d: %s", filename, line, buffer);
         }

         if(buffer[y] != string_to_search[u])
         {
            u = 0;
            y++;
         }
      }
   }

   fclose(file);
   return 0;
}

void *threadsearch(void *arg)
{
   /* threaded search */

   char *filename;
   char *string_to_search;

   thread_data *my_data = (thread_data *)arg;
   
   filename = my_data->filename;
   string_to_search = my_data->string_to_search;

   searchstring(filename, string_to_search);

   return ((void *)NULL);
}
